
#include "losing.h"

// bit 31 of data set, has trans
// bit 30 of data set, has child, or has sibling when has trans
// 10 trans, no sibling, data is trans
// 01 has child (+1), data is sibling (or zero)
// 11 trans and sibling (+1), data is trans
// if data is 0x3fffffff, then UNWON, if (1U<<30) is set, has sibling

#define SET_CHILD(A,x) (A[x].data|=(1U<<30))
#define DATA_IS_TRANS(A,x) (((A[x].data&(1U<<31)))==(1U<<31))
#define DATA_HAS_CHILD(A,x)\
  (((A[x].data&(3U<<30)))==(1U<<30) && ((A[x].data&0x3fffffff)!=0x3fffffff))
#define TRANS_AND_SIBLING(A,x) (((A[x].data&(3U<<30)))==(3U<<30))
#define DATA_GET_NODE(A,x) (A[x].data&0x3fffffff)
#define DATA_GET_NODEns(A,x) \
  (((A[x].data&0x3fffffff)!=0x3fffffff) ? A[x].data&0x3fffffff : \
   (((A[x].data&(1U<<30))==(1U<<30)) ? ((x)+1) : 0))
#define NEXT_SIBLING(A,x) ((TRANS_AND_SIBLING(A,x)) ? ((x)+1) : \
			   (DATA_IS_TRANS(A,x) ? 0 : DATA_GET_NODEns(A,x)))

static uint32 subtree_size(WIN_NODE *W,uint32 n,boolean t);

static uint64 *ARR;
#define GET_ARRAY_BIT(A,x) (A[(x)>>6]&(1ULL<<((x)&0x3f)))
#define SET_ARRAY_BIT(A,x) A[(x)>>6]|=(1ULL<<((x)&0x3f))

static void walk_tree(WIN_NODE *W,uint32 n,boolean t)
{uint32 c; if (t) {if (GET_ARRAY_BIT(ARR,n)) return; SET_ARRAY_BIT(ARR,n);}
 if (DATA_IS_TRANS(W,n)) {walk_tree(W,DATA_GET_NODE(W,n),t); return;}
 if (!t) {if (GET_ARRAY_BIT(ARR,n)) return; SET_ARRAY_BIT(ARR,n);}
 if (!DATA_HAS_CHILD(W,n)) return; c=n+1; walk_tree(W,c,t);
 while ((c=NEXT_SIBLING(W,c))) walk_tree(W,c,t);}

static uint32 subtree_size(WIN_NODE *W,uint32 n,boolean t)
{uint32 u,ct=0,sz=W[0].data&0x3fffffff; if (!n) return sz; sz=(sz+63)/64;
 if (DATA_GET_NODE(W,n)==0x3fffffff) return -1;
 while (DATA_IS_TRANS(W,n)) n=DATA_GET_NODE(W,n);
 for (u=0;u<sz;u++) ARR[u]=0; walk_tree(W,n,t);
 for (u=0;u<sz;u++) ct+=POPCNT(ARR[u]); return ct;}

typedef struct {sint32 size; uint32 n; uint16 move;} tRECORD;
static int compare(const void *x,const void *y)
{tRECORD a=*((tRECORD*) (x)); tRECORD b=*((tRECORD*) (y));
 if (a.size<0 && b.size>=0) return -1; if (b.size<0 && a.size>=0) return 1;
 if (ABS(a.size)>ABS(b.size)) return -1;
 if (ABS(a.size)<ABS(b.size)) return 1; return 0;}

static void output_children(WIN_NODE *W,uint32 n,boolean t)
{uint32 z,u,c,r=0; sint32 sz; tRECORD REC[256]; char N1[8],N2[8]; c=n+1;
 REC[r].n=c; REC[r].move=W[c].move; REC[r++].size=subtree_size(W,c,t);
 while ((c=NEXT_SIBLING(W,c)))
 {REC[r].n=c; REC[r].move=W[c].move; REC[r++].size=subtree_size(W,c,t);}
 qsort(REC,r,sizeof(tRECORD),compare); sz=subtree_size(W,n,t);
 printf("Val: KNOWN SIZE %d\n",sz);
 for (u=0;u<r;u++)
 {z=DATA_IS_TRANS(W,REC[u].n)?DATA_GET_NODE(W,REC[u].n):REC[u].n;
  printf("%s KNOWN %s %d* RES 0\n",Notate(N1,REC[u].move),
	 DATA_HAS_CHILD(W,z) && DATA_GET_NODE(W,z)!=0x3fffffff
	 && !NEXT_SIBLING(W,z+1) ? Notate(N2,W[z+1].move) : "---",
	 ABS(REC[u].size));}}

static uint32 tree_move(WIN_NODE *W,uint16 mv,uint32 n)
{uint32 c;
 if (!DATA_HAS_CHILD(W,n)) return -1; c=n+1; if (W[c].move==mv) return c;
 while ((c=NEXT_SIBLING(W,c))) if (W[c].move==mv) return c; return -1;}

static uint32 PROLOG_LEN;
static uint32 align_node(WIN_NODE *W,typePOS *POS)
{uint32 n=0; typeDYNAMIC *D; D=POS->DYN_ROOT+PROLOG_LEN+1;
 while (D!=POS->DYN && n!=0xffffffff)
 {D++; n=tree_move(W,D->mv,n);
  while (DATA_IS_TRANS(W,n)) n=DATA_GET_NODE(W,n);} return n;}

static uint64 PROLOG_HASH;
static boolean COUNT_TRANS=FALSE;
static void DoQuery(WIN_NODE *W,typePOS *POS)
{uint32 n=0; n=align_node(W,POS);
 if (!n && POS->DYN->HASH!=PROLOG_HASH) goto NONE;
 if (DATA_HAS_CHILD(W,n)) {output_children(W,n,COUNT_TRANS); return;}
 printf("No children %d\n",0); return; NONE: printf("Not in BOOK\n");}

uint32 follow_known(typePOS *POS,typeDYNAMIC *D)
{int i; WIN_NODE *W; WIN_STRUCT *WS; uint64 sz; typePOS POS2[1];
 memset((void*) POS2,0,sizeof(typePOS));
 POS2->DYN_ROOT=CALLOC(MAX_PLY,sizeof(typeDYNAMIC));
 POS2->DYN=POS2->DYN_ROOT+1; init_position(POS2,"startpos");
 // WS=LoadWintreeFile(get_name(D->HASH,KNOWN_NAME),TRUE);
 WS=WintreeKnown(D->HASH);
 if (!WS) {printf("follow_known() fail\n"); fflush(stdout); return 0xffffffff;}
 PROLOG_HASH=D->HASH; W=WS->W; PROLOG_LEN=W[0].move;
 sz=W[0].data&0x3fffffff; ARR=malloc(sz/8+64); SET_CHILD(W,0);
 for (i=0;i<PROLOG_LEN;i++) MakeMove(POS2,WS->MOVE_LIST[i]); D++;
 while (D!=POS->DYN+1) {MakeMove(POS2,D->mv); D++;} DoQuery(W,POS2);
 free(W); free(WS->MOVE_LIST); free(WS); free(ARR);
 printf("Book Query Complete\n"); fflush(stdout); return 0xffffffff;}
