
#include "losing.h"

// bit 31 of data set, has trans
// bit 30 of data set, has child, or has sibling when has trans
// 10 trans, no sibling, data is trans
// 01 has child (+1), data is sibling (or zero)
// 11 trans and sibling (+1), data is trans

#define DATA_IS_TRANS(A,x) (((A[x].data&(1U<<31)))==(1U<<31))
#define DATA_HAS_CHILD(A,x) (((A[x].data&(3U<<30)))==(1U<<30))
#define TRANS_AND_SIBLING(A,x) (((A[x].data&(3U<<30)))==(3U<<30))
#define DATA_GET_NODE(A,x) (A[x].data&0x3fffffff)
#define SET_CHILD(A,x) (A[x].data|=(1U<<30))
#define NEXT_SIBLING(A,x) ((TRANS_AND_SIBLING(A,x)) ? ((x)+1) : \
			   (DATA_IS_TRANS(A,x) ? 0 : DATA_GET_NODE(A,x)))

static uint32 *LEV2_CHANGE; static WIN_NODE *C; static uint64 SIZE=0;

static void resize()
{uint64 ns=SIZE+SIZE/4; int u; printf("Resize wintree %lld to %lld\n",SIZE,ns);
 C=realloc(C,ns*sizeof(WIN_NODE));
 for (u=SIZE;u<ns;u++) C[u].data=C[u].move=0; SIZE=ns;}

static int is_unsolved(uint16 mv)
{if ((mv!=0xfedc) && ((mv>>12)>8 || (mv>>12)==7)) return TRUE; return FALSE;}

static uint32 CopyWithMissing(WIN_NODE *U,uint32 n,uint32 c) // C is global
{uint32 last=c; boolean adv=FALSE; if (c>SIZE-256) resize();
 if (is_unsolved(U[n].move)) goto LOOP; LEV2_CHANGE[n]=c; C[c].move=U[n].move;
 if (DATA_IS_TRANS(U,n)) C[c].data=(1U<<31)|LEV2_CHANGE[DATA_GET_NODE(U,n)];
 if (DATA_HAS_CHILD(U,n)) {SET_CHILD(C,c); c=CopyWithMissing(U,n+1,++c);}
 else c++; adv=TRUE;
LOOP:
 while ((n=NEXT_SIBLING(U,n)))
 {if (is_unsolved(U[n].move)) continue; LEV2_CHANGE[n]=c; C[c].move=U[n].move;
  if (adv && DATA_IS_TRANS(C,last)) C[last].data|=(1U<<30); // TRANS+SIBLING
  else if (adv) C[last].data|=c; last=c; adv=TRUE;
  if (DATA_IS_TRANS(U,n))
  {C[c++].data=(1U<<31)|LEV2_CHANGE[DATA_GET_NODE(U,n)]; continue;}
  if (DATA_HAS_CHILD(U,n)) {SET_CHILD(C,c); c=CopyWithMissing(U,n+1,++c);}
  else c++;} return c;}

static int comp(const void *x,const void *y)
{uint16 a=*((uint16*) x); uint16 b=*((uint16*) y);
 if (a>b) return 1; if (a<b) return -1; printf("BADSORT %o\n",a); return 0;}

static uint32 CopyWithVerify(typePOS *POS,WIN_NODE *U,uint32 n,uint32 c)
{uint32 f=c,last=c; boolean adv=FALSE; int wc=0,wg=0; uint16 MC[256],MG[256];
 int t=0,u; if (c>SIZE-256) resize(); if (is_unsolved(U[n].move)) goto LOOP;
 LEV2_CHANGE[n]=c; C[c].move=U[n].move;
 if (DATA_IS_TRANS(U,n)) C[c].data=(1U<<31)|LEV2_CHANGE[DATA_GET_NODE(U,n)];
 if (DATA_HAS_CHILD(U,n))
 {SET_CHILD(C,c); MakeMoveCheck(POS,U[n].move,FALSE);
  c=CopyWithVerify(POS,U,n+1,++c); UnmakeMove(POS,U[n].move,TRUE);}
 else c++; adv=TRUE;
LOOP:
 while ((n=NEXT_SIBLING(U,n)))
 {if (is_unsolved(U[n].move)) continue; LEV2_CHANGE[n]=c; C[c].move=U[n].move;
  if (adv && DATA_IS_TRANS(C,last)) C[last].data|=(1U<<30); // TRANS+SIBLING
  else if (adv) C[last].data|=c; last=c; adv=TRUE;
  if (DATA_IS_TRANS(U,n))
  {C[c++].data=(1U<<31)|LEV2_CHANGE[DATA_GET_NODE(U,n)]; continue;}
  if (DATA_HAS_CHILD(U,n))
  {SET_CHILD(C,c); MakeMoveCheck(POS,U[n].move,FALSE);
   c=CopyWithVerify(POS,U,n+1,++c); UnmakeMove(POS,U[n].move,TRUE);}
  else c++;}
 // if btm, starting at f, collect all moves, compare to list from movegen
 if (POS->wtm) return c;
 do {MC[wc++]=C[f].move;} while ((f=NEXT_SIBLING(C,f)));
 wg=GenMoves(POS,MG); if (wg==wc) return c;
 qsort(MC,wc,sizeof(uint16),comp); qsort(MG,wg,sizeof(uint16),comp);
 dump_path(POS); printf("Have %d with %d possible\n",wc,wg); // raise(SIGTRAP);
 for (u=0;u<wg;u++)
 {if (t<wc && MG[u]==MC[t]) {t++; continue;} C[c].move=MG[u]^0xf000;
  if (DATA_IS_TRANS(C,last)) C[last].data|=(1U<<30); else C[last].data|=c;
  last=c; C[c++].data=0;} return c;}

static uint32 CopyWinTree(WIN_NODE *U,uint32 n,uint32 c) // C is global
{uint32 last=c; if (c>SIZE-256) resize();
 LEV2_CHANGE[n]=c; C[c].move=U[n].move;
 if (DATA_IS_TRANS(U,n)) C[c].data=(1U<<31)|LEV2_CHANGE[DATA_GET_NODE(U,n)];
 if (DATA_HAS_CHILD(U,n)) {SET_CHILD(C,c); c=CopyWinTree(U,n+1,++c);} else c++;
 while ((n=NEXT_SIBLING(U,n)))
 {LEV2_CHANGE[n]=c; C[c].move=U[n].move;
  if (DATA_IS_TRANS(C,last)) C[last].data|=(1U<<30); // TRANS_AND_SIBLING
  else C[last].data|=c; last=c;
  if (DATA_IS_TRANS(U,n))
  {C[c++].data=(1U<<31)|LEV2_CHANGE[DATA_GET_NODE(U,n)]; continue;}
  if (DATA_HAS_CHILD(U,n)) {SET_CHILD(C,c); c=CopyWinTree(U,n+1,++c);}
  else c++;} return c;}

////////////////////////////////////////////////////////////////////////

static int get_wintree_prolog(typePOS *POS,WIN_NODE *TREE,uint16 *ML)
{uint32 c=1,w=0;
 while (!NEXT_SIBLING(TREE,c))
 {MakeMove(POS,TREE[c].move); ML[w++]=TREE[c].move; c++;} return w;}

static uint16 PROLOG[256]; // PROLOG here is hackish
static WIN_NODE* LoadFile(char *A) // should be biffed, really
{FILE *F; char B[1024]; WIN_NODE *W; uint64 u,d,s; // 64-bit (2GB)
 printf("Loading win-tree %s\n",A); W=malloc(32); F=fopen(A,"rb");
 if (!F) {strcpy(B,"PROOFS/"); strcat(B,A); F=fopen(B,"rb");}
 if (!F) {printf("File %s not found\n",A); exit(1);}
 fread(&(W[0].data),4,1,F); fread(&(W[0].move),2,1,F);
 for (u=0;u<W[0].move;u++) fread(&(PROLOG[u]),2,1,F);
 d=W[0].data&0x3fffffff; s=(11*d)/10+0x10000; W=realloc(W,s*sizeof(WIN_NODE));
 for (u=1;u<d;u++) {fread(&(W[u].data),4,1,F); fread(&(W[u].move),2,1,F);}
 for (u=d;u<s;u++) {W[u].data=W[u].move=0;} fclose(F);
 printf("Loaded %lld records [expansion up to %lld]\n",d,s); return W;}

////////////////////////////////////////////////////////////////////////

static void do_missing(char *fin,char *fout,boolean SAVE)
{typePOS POS[1]; uint32 sz,u; WIN_NODE *W;
 magic_mult_init(); POS->DYN_ROOT=CALLOC(MAX_PLY,sizeof(typeDYNAMIC));
 POS->DYN=POS->DYN_ROOT+1; W=LoadFile(fin); // a bit hackish with PROLOG
 init_position(POS,"startpos"); printf ("MISSING WIN-TREE\n"); fflush(stdout);
 sz=(W[0].data&0x3fffffff)+1024; C=malloc(sz*sizeof(WIN_NODE));
 LEV2_CHANGE=malloc((sz+1024)*sizeof(uint32)); SIZE=sz; // should be enough
 for (u=0;u<sz;u++) C[u].data=C[u].move=0; SET_CHILD(C,0);
 C[0].data=(1U<<30)|CopyWithMissing(W,1,1); C[0].move=W[0].move;
 if (SAVE) SaveWinTree(C,PROLOG,C[0].move,fout);}

static void do_unsolved(char *fin,char *fout,boolean SAVE)
{typePOS POS[1]; uint32 sz,u; WIN_NODE *W; magic_mult_init();
 REV_LIMIT=125; POS->DYN_ROOT=CALLOC(MAX_PLY,sizeof(typeDYNAMIC));
 POS->DYN=POS->DYN_ROOT+1; W=LoadFile(fin); // a bit hackish with PROLOG
 init_position(POS,"startpos"); printf ("UNSOLVED WIN-TREE\n"); fflush(stdout);
 sz=(W[0].data&0x3fffffff)+1024; C=malloc(sz*sizeof(WIN_NODE));
 LEV2_CHANGE=malloc((sz+1024)*sizeof(uint32)); SIZE=sz; // should be enough
 for (u=0;u<sz;u++) C[u].data=C[u].move=0; SET_CHILD(C,0);
 for (u=0;u<W[0].move;u++) MakeMoveCheck(POS,PROLOG[u],FALSE);
 C[0].data=(1U<<30)|CopyWithVerify(POS,W,1,1); C[0].move=W[0].move;
 if (SAVE) SaveWinTree(C,PROLOG,C[0].move,fout);}

static void do_prolog(char *fin,char *fout,boolean SAVE)
{typePOS POS[1]; WIN_NODE *TREE; uint32 w,u,sz;
 char N[8]; uint16 ML[1024]; magic_mult_init();
 POS->DYN_ROOT=CALLOC(MAX_PLY,sizeof(typeDYNAMIC)); POS->DYN=POS->DYN_ROOT+1;
 TREE=LoadFile(fin); if (TREE[0].move) {printf("Prolog already\n"); exit(1);}
 init_position(POS,"startpos"); printf ("PROLOG WIN-TREE\n"); fflush(stdout);
 w=get_wintree_prolog(POS,TREE,ML); printf("Prolog:");
 for (u=0;u<w;u++) printf(" %s",Notate(N,ML[u])); printf("\n");
 for (u=0;u<w;u++) MakeMove(POS,ML[u]); sz=(TREE[0].data&0x3fffffff);
 LEV2_CHANGE=malloc((sz+1024+w)*sizeof(uint32)); SIZE=sz+1024+w;
 C=malloc((sz+1024+w)*sizeof(WIN_NODE)); // should be enough
 for(u=0;u<sz;u++) C[u].data=C[u].move=0; SET_CHILD(C,0);
 C[0].data=CopyWinTree(TREE,w,0)|(1<<30); if (SAVE) SaveWinTree(C,ML,w,fout);}

static void do_serial(char *fin,char *fout,boolean SAVE)
{typePOS POS[1]; uint32 sz,u; WIN_NODE *W;
 magic_mult_init(); POS->DYN_ROOT=CALLOC(MAX_PLY,sizeof(typeDYNAMIC));
 POS->DYN=POS->DYN_ROOT+1; W=LoadFile(fin); // a bit hackish with PROLOG
 init_position(POS,"startpos"); printf ("SERIAL WIN-TREE\n"); fflush(stdout);
 sz=W[0].move+(W[0].data&0x3fffffff)+1024; C=malloc(sz*sizeof(WIN_NODE));
 LEV2_CHANGE=malloc((sz+1024)*sizeof(uint32)); SIZE=sz; // should be enough
 for(u=0;u<sz;u++) C[u].data=C[u].move=0; SET_CHILD(C,0);
 for (u=0;u<W[0].move;u++)
 {MakeMove(POS,PROLOG[u]); SET_CHILD(C,u+1); C[u+1].move=PROLOG[u];}
 if ((W[0].data&0x3fffffff)>1) C[0].data=(1U<<30)|CopyWinTree(W,1,1+W[0].move);
 else {C[W[0].move].data=0; C[0].data=(1U<<30)|(W[0].move+1);}
 if (SAVE) SaveWinTree(C,NULL,0,fout);}

////////////////////////////////////////////////////////////////////////

void expand_and_rehash(WIN_STRUCT *WS,char *fout,boolean SAVE);

static void do_expand(char *fin,char *fout,boolean SAVE)
{typePOS POS[1]; WIN_STRUCT *WS;
 memset((void*) POS,0,sizeof(typePOS)); magic_mult_init(); init_known(FALSE);
 POS->DYN_ROOT=CALLOC(MAX_PLY,sizeof(typeDYNAMIC)); POS->DYN=POS->DYN_ROOT+1;
 init_position(POS,"startpos"); printf("WINTREE EXPAND\n"); fflush(stdout);
 WS=LoadWintreeFile(fin,FALSE); expand_and_rehash(WS,fout,SAVE);}

void newrehash(WIN_NODE *WIN,uint16 *ML,int w,char *fn,boolean SAVE);

static void do_rehash(char *fin,char *fout,boolean SAVE)
{typePOS POS[1]; WIN_STRUCT *WS;
 memset((void*) POS,0,sizeof(typePOS)); magic_mult_init();
 POS->DYN_ROOT=CALLOC(MAX_PLY,sizeof(typeDYNAMIC)); POS->DYN=POS->DYN_ROOT+1;
 init_position(POS,"startpos"); printf("WINTREE REHASH\n"); fflush(stdout);
 WS=LoadWintreeFile(fin,FALSE);
 newrehash(WS->W,WS->MOVE_LIST,WS->W[0].move,fout,SAVE);}

void do_splice(typePOS *POS,WIN_NODE *W1,WIN_NODE *W2,boolean SAVE);

static void setup_splice(char *fn1,char *fn2) // could be more verbose
{WIN_NODE *W1,*W2; typePOS POS[1]; // maybe should rehash first?
 POS->DYN_ROOT=CALLOC(MAX_PLY,sizeof(typeDYNAMIC)); POS->DYN=POS->DYN_ROOT+1;
 do_serial(fn1,NULL,FALSE); W1=C; do_serial(fn2,NULL,FALSE); W2=C;
 // note that do_serial() is called twice, so magic_mult_init() is also
 REV_LIMIT=125; // would rather not do this, as resulting tree is bad
 do_splice(POS,W1,W2,TRUE);}

////////////////////////////////////////////////////////////////////////

static uint32 WinTreeMove(WIN_NODE *W,uint16 mv,uint32 n)
{uint32 c=n+1; if (!DATA_HAS_CHILD(W,n)) return -1;
 do {if (W[c].move==mv) return DATA_IS_TRANS(W,c)?DATA_GET_NODE(W,c):c;}
 while ((c=NEXT_SIBLING(W,c))); return -1;}

typedef struct {uint64 key; uint32 main,t;} tTRANS2;
static tTRANS2 *HASH_TABLE;
#define HASH_MASKING (HASH_LIMIT-1)
static uint32 HASH_LIMIT=0;

static uint32 HASH_LOOKUP(uint64 Z)
{uint64 h; uint32 k=Z&HASH_MASKING; if (!k) k++;
 while ((h=HASH_TABLE[k].key))
 {if (h==Z) return k; k++; if (k==HASH_LIMIT) k=1;} return 0;}

static uint32 INCLUDE_HASH(uint64 Z,uint32 n)
{uint64 h; uint32 k=Z&HASH_MASKING; if (!k) k++;
 while ((h=HASH_TABLE[k].key))
 {if (h!=Z) {k++; if (k==HASH_LIMIT) k=1; continue;}
  if (HASH_TABLE[k].t!=-1) return TRUE; break;}
 HASH_TABLE[k].key=Z; HASH_TABLE[k].t=n; return FALSE;}

static uint16 UU(uint16 mv) // 0,2,3,4,5,6,8 -> // 15,13,12,11,10,9,7
{uint16 k=mv>>12; if (mv==0xfedc) return mv;
 if (k>8 || k==7) return mv^0xf000; return mv;}

static void fill_hash_table(typePOS *POS,WIN_NODE *W,uint32 n)
{uint16 mv; // this requires that HASH is tree-unique [can be done via rehash]
 if (INCLUDE_HASH(POS->DYN->HASH,n) || !DATA_HAS_CHILD(W,n)) return;
 n++; mv=UU(W[n].move);
 MakeMove(POS,mv); fill_hash_table(POS,W,n); UnmakeMove(POS,mv,0);
 while ((n=NEXT_SIBLING(W,n)))
 {mv=UU(W[n].move); MakeMove(POS,mv);
  fill_hash_table(POS,W,n); UnmakeMove(POS,mv,0);} return;}

static uint32 StepWinTree
(typePOS *POS,WIN_NODE *W,WIN_NODE *NEW,uint32 NODE,uint64 H)
{uint32 c,h,n,LAST; uint16 G[256],mv; uint32 w=1; G[1]=0;
 if (H && H==POS->DYN->HASH) {NEW[NODE].move^=0xf000; return NODE+1;}
 h=HASH_LOOKUP(POS->DYN->HASH); if (!h) printf("ERROR\n");
 if (HASH_TABLE[h].main)
 {NEW[NODE++].data=(1U<<31)|HASH_TABLE[h].main; return NODE;}
 HASH_TABLE[h].main=NODE; n=HASH_TABLE[h].t;
 if ((n==-1 || !DATA_HAS_CHILD(W,n))) return NODE+1;
 c=n+1; mv=W[c].move;
 while (mv!=UU(mv)) {c=NEXT_SIBLING(W,c); mv=W[c].move;}
 G[0]=mv;
 while ((c=NEXT_SIBLING(W,c)))
 {mv=W[c].move; if (mv!=UU(mv)) continue; G[w++]=mv;}
 G[w]=0; w=0; mv=G[w++]; SET_CHILD(NEW,NODE); NEW[++NODE].move=mv; LAST=NODE;
 MakeMove(POS,mv); NODE=StepWinTree(POS,W,NEW,NODE,H); UnmakeMove(POS,mv,0);
 while ((mv=G[w++]))
 {if (NEW[LAST].data&(1U<<31)) NEW[LAST].data|=(1U<<30);
  else NEW[LAST].data|=NODE; NEW[NODE].move=mv; LAST=NODE;
  MakeMove(POS,mv); NODE=StepWinTree(POS,W,NEW,NODE,H);
  UnmakeMove(POS,mv,0);} return NODE;}

static void do_subexclude(char *fin,char *fout,boolean SAVE)
{WIN_NODE *W,*NEW; typePOS POS[1]; uint32 w=0,n=0; uint64 u,sz,asz,H;
 magic_mult_init(); POS->DYN_ROOT=CALLOC(MAX_PLY,sizeof(typeDYNAMIC));
 POS->DYN=POS->DYN_ROOT+1; do_serial(fin,NULL,FALSE); // MakeMove
 init_position(POS,"startpos"); REV_LIMIT=125; W=C; sz=W[0].data&0x3fffffff;
 HASH_LIMIT=2U<<BSR(sz); asz=((uint64) HASH_LIMIT)*sizeof(tTRANS2);
 printf("Allocating hash table %lldmb\n",asz>>20); HASH_TABLE=malloc(asz);   
 for (u=0;u<HASH_LIMIT;u++) HASH_TABLE[u].key=HASH_TABLE[u].main=0;   
 for (u=0;u<HASH_LIMIT;u++) HASH_TABLE[u].t=-1; fill_hash_table(POS,W,0);
 printf("Parser ready\n"); char A[4096]; fgets(A,4000,stdin); A[strlen(A)-1]=0;
 if (memcmp(A,"moves ",6)) return;
 parse_moves(POS,A+6); typeDYNAMIC *D; D=POS->DYN_ROOT+1;
 while (D!=POS->DYN && n!=-1) {D++; n=WinTreeMove(W,D->mv,n);}
 if (n==-1) {printf("Failure to align moves\n"); return;}
 H=POS->DYN->HASH; UnmakeMove(POS,D->mv,0); printf("Exclude hash 0x%llx\n",H);
 uint16 ML[1024]; D=POS->DYN_ROOT+1; while (D!=POS->DYN) {D++; ML[w++]=D->mv;}
 asz=sz*sizeof(WIN_NODE); printf("Alloc %lldmb for new tree\n",asz>>20);
 NEW=malloc(asz); for (u=0;u<sz;u++) NEW[u].data=NEW[u].move=0;
 NEW[0].data=(1U<<30)|StepWinTree(POS,W,NEW,0,H);
 printf("New size is %d\n",DATA_GET_NODE(NEW,0));
 if (SAVE) SaveWinTree(NEW,ML,w,fout);}

static void do_subtree(char *fin,char *fout,boolean SAVE)
{WIN_NODE *W,*NEW; typePOS POS[1]; uint32 w=0,n=0; uint64 u,sz,asz;
 magic_mult_init(); POS->DYN_ROOT=CALLOC(MAX_PLY,sizeof(typeDYNAMIC));
 POS->DYN=POS->DYN_ROOT+1; do_serial(fin,NULL,FALSE); // MakeMove
 init_position(POS,"startpos"); REV_LIMIT=125; W=C; sz=W[0].data&0x3fffffff;
 HASH_LIMIT=2U<<BSR(sz); asz=((uint64) HASH_LIMIT)*sizeof(tTRANS2);
 printf("Allocating hash table %lldmb\n",asz>>20); HASH_TABLE=malloc(asz);   
 for (u=0;u<HASH_LIMIT;u++) HASH_TABLE[u].key=HASH_TABLE[u].main=0;   
 for (u=0;u<HASH_LIMIT;u++) HASH_TABLE[u].t=-1; fill_hash_table(POS,W,0);
 printf("Parser ready\n"); char A[4096]; fgets(A,4000,stdin); A[strlen(A)-1]=0;
 if (memcmp(A,"moves ",6)) return;
 parse_moves(POS,A+6); typeDYNAMIC *D; D=POS->DYN_ROOT+1;
 while (D!=POS->DYN && n!=-1) {D++; n=WinTreeMove(W,D->mv,n);}
 if (n==-1) {printf("Failure to align moves\n"); return;}
 uint16 ML[1024]; D=POS->DYN_ROOT+1; while (D!=POS->DYN) {D++; ML[w++]=D->mv;}
 asz=sz*sizeof(WIN_NODE); printf("Alloc %lldmb for new tree\n",asz>>20);
 NEW=malloc(asz); for (u=0;u<sz;u++) NEW[u].data=NEW[u].move=0;
 NEW[0].data=(1U<<30)|StepWinTree(POS,W,NEW,0,0);
 printf("New size is %d\n",DATA_GET_NODE(NEW,0));
 if (SAVE) SaveWinTree(NEW,ML,w,fout);}

////////////////////////////////////////////////////////////////////////

#define MILLION (0x100000)
#define COUNTED_SIZE (128*MILLION)
#define COUNTED_MASK (COUNTED_SIZE-1)

static uint64 *ARR;
#define GET_ARRAY_BIT(A,x) (A[(x)>>6]&(1ULL<<((x)&0x3f)))
#define SET_ARRAY_BIT(A,x) A[(x)>>6]|=(1ULL<<((x)&0x3f))

static boolean SOLVED;
static void walk_tree(WIN_NODE *W,uint32 n) // ,boolean t) // t=FALSE
{uint32 c;
 if (DATA_GET_NODE(W,n)==0x3fffffff || is_unsolved(W[n].move)) SOLVED=FALSE;
 if (DATA_GET_NODE(W,n)==0x3fffffff || is_unsolved(W[n].move)) return;
 if (DATA_IS_TRANS(W,n)) {walk_tree(W,DATA_GET_NODE(W,n)); return;}
 if (TRUE) {if (GET_ARRAY_BIT(ARR,n)) return; SET_ARRAY_BIT(ARR,n);}
 if (!DATA_HAS_CHILD(W,n)) return; c=n+1; walk_tree(W,c);
 while ((c=NEXT_SIBLING(W,c))) walk_tree(W,c);}

static uint32 COMPUTE_HASH(uint32 n) /* totally random (the method that is) */
{uint32 k=n*n; k+=(n>>10)^((~n)&0x3ff); k+=((n>>5)^((~n)&0x7fff))<<5;
  k+=(((~n)>>15)^(n&0x1f))<<5; k+=(n>>4)&0x55aa55; k+=((~n)>>8)&0xaa55aa;
  return k&COUNTED_MASK;}

#define SIGN(x) (((x)>0)?1:-1)
typedef struct {uint32 n; sint32 sz;} tCOUNTED;
static tCOUNTED *COUNTED;
static sint32 SUBTREE_SIZE_IS_COUNTED(WIN_NODE *W,uint32 n)
{uint32 k; while (DATA_IS_TRANS(W,n)) n=DATA_GET_NODE(W,n); k=COMPUTE_HASH(n);
 while (COUNTED[k].n)
 {if (n==COUNTED[k].n) return COUNTED[k].sz; k++; if (k==COUNTED_SIZE) k=0;}
  return 0;}
static void SAVE_SUBTREE_SIZE(WIN_NODE *W,uint32 n,sint32 ct)
{uint32 k; while (DATA_IS_TRANS(W,n)) n=DATA_GET_NODE(W,n);
 k=COMPUTE_HASH(n); while (COUNTED[k].n) {k++; if (k==COUNTED_SIZE) k=0;}
 COUNTED[k].sz=ct; COUNTED[k].n=n; if (!DATA_HAS_CHILD(W,n)) return;}

static uint32 subtree_size(WIN_NODE *W,uint32 n)
{uint32 s,u,ct=0,sz=W[0].data&0x3fffffff; if (!n) return sz; sz=(sz+63)/64;
 if (DATA_GET_NODE(W,n)==0x3fffffff || is_unsolved(W[n].move)) return -1;
 while (DATA_IS_TRANS(W,n)) n=DATA_GET_NODE(W,n);
 if ((s=SUBTREE_SIZE_IS_COUNTED(W,n))) return s; SOLVED=TRUE;
 for (u=0;u<sz;u++) ARR[u]=0; walk_tree(W,n);
 for (u=0;u<sz;u++) ct+=POPCNT(ARR[u]);
 if (!SOLVED) ct=-ct; SAVE_SUBTREE_SIZE(W,n,ct); return ct;}

static tCOUNTED *RECORD; static uint32 rec=0;
#define record_it(a,b) {RECORD[rec].n=a; RECORD[rec].sz=b; rec++;\
                        printf("Record %d: node %d size %d\n",rec,a,b);}
static void size_loop(WIN_NODE *W,uint32 n,uint32 lim)
{uint32 c,sz; if (DATA_IS_TRANS(W,n)) n=DATA_GET_NODE(W,n);
 if (SUBTREE_SIZE_IS_COUNTED(W,n)) return;
 sz=subtree_size(W,n); if (sz<lim) return; record_it(n,sz);
 c=n+1; size_loop(W,c,lim); while ((c=NEXT_SIBLING(W,c))) size_loop(W,c,lim);}

static void do_precount(int lim,char *fn) // needs full filename
{WIN_STRUCT *WS; int i,u; uint32 sz; FILE *F; char A[1024];
 WS=LoadWintreeFile(fn,FALSE); COUNTED=malloc(COUNTED_SIZE*sizeof(tCOUNTED));
 RECORD=malloc(MILLION*sizeof(tCOUNTED)); // probably MILLION is enough
 for (i=0;i<COUNTED_SIZE;i++) COUNTED[i].n=COUNTED[i].sz=0;
 sz=WS->W[0].data&0x3fffffff; ARR=malloc(sz/8+64); size_loop(WS->W,0,lim);
 strcpy(A,fn); strcat(A,".copy"); F=fopen(A,"w");
 if (!F) {printf("File open failed\n"); return;}
 fwrite(&(WS->W[0].data),4,1,F); fwrite(&(WS->W[0].move),2,1,F);
 for (u=0;u<WS->W[0].move;u++) fwrite(&(WS->MOVE_LIST[u]),2,1,F);
 for (u=1;u<((WS->W[0].data)&0x3fffffff);u++)
 {fwrite(&(WS->W[u].data),4,1,F); fwrite(&(WS->W[u].move),2,1,F);}
 for (u=0;u<rec;u++) {fwrite(&RECORD[u].n,4,1,F); fwrite(&RECORD[u].sz,4,1,F);}
 printf("Saved new precount information to %s\n",fn); fclose(F); rename(A,fn);}

////////////////////////////////////////////////////////////////////////

int main(int c,char **I)
{char STR[256]; if (c>3) strcpy(STR,I[3]);
 if (!strcmp(I[1],"serial"))
 {if (c<=3) strcpy(STR,"PROOFS/SERIAL.proof"); do_serial(I[2],STR,TRUE);}
 if (!strcmp(I[1],"prolog"))
 {if (c<=3) strcpy(STR,"PROOFS/PROLOG.proof"); do_prolog(I[2],STR,TRUE);}
 if (!strcmp(I[1],"missing"))
 {if (c<=3) strcpy(STR,"PROOFS/MISSING.proof"); do_missing(I[2],STR,TRUE);}
 if (!strcmp(I[1],"unsolved"))
 {if (c<=3) strcpy(STR,"PROOFS/UNSOLVED.proof"); do_unsolved(I[2],STR,TRUE);}
 if (!strcmp(I[1],"expand"))
 {if (c<=3) strcpy(STR,"PROOFS/EXPANDED.proof"); do_expand(I[2],STR,TRUE);}
 if (!strcmp(I[1],"rehash"))
 {if (c<=3) strcpy(STR,"PROOFS/REHASH.proof"); do_rehash(I[2],STR,TRUE);}
 if (!strcmp(I[1],"subtree"))
 {if (c<=3) strcpy(STR,"PROOFS/SUBTREE.proof"); do_subtree(I[2],STR,TRUE);}
 if (!strcmp(I[1],"subexclude"))
 {if (c<=3) strcpy(STR,"PROOFS/SUBEXCLUDE.proof");
  do_subexclude(I[2],STR,TRUE);}
 if (!strcmp(I[1],"splice")) setup_splice(I[2],I[3]);
 if (!strcmp(I[1],"precount")) do_precount(atoi(I[2]),I[3]);
 return 0;}
