
#include "losing.h"
#include <math.h>

////////////////////////////////////////////////////////////////////////

uint32 follow_known(typePOS *POS,typeDYNAMIC *D); // cluster_known.c

static uint32 iTreeMove(PN_NODE *T,uint16 mv,uint32 n)
{uint32 c=T[n].child; if (!n) return 0;
 do {if (T[c].move==mv) return c;} while ((c=T[c].sibling)); return 0;}

static uint32 TreeMove(PN_NODE *T,uint16 mv,uint32 n)
{uint32 r=iTreeMove(T,mv,n); if (r==0) return 0;
 if (T[r].loop || T[r].flags&4) return T[r].trans; return r;}

typedef struct {uint32 n,good,bad;} SORTER;
static int sort_compare(const void *x, const void *y) // increasing good
{SORTER A=*((SORTER*) x); SORTER B=*((SORTER*) y);
  if (A.good<B.good) return -1; if (A.good>B.good) return 1;
  if (A.bad>B.bad) return -1; if (A.bad<B.bad) return 1; return 0;}

static void SortChildren(PN_NODE *T,uint32 n)
{SORTER S[256]; int i,u=0; uint32 c=T[n].child;
 S[u].bad=T[c].bad; S[u].good=T[c].good; S[u++].n=c;
 while (T[c].sibling)
 {c=T[c].sibling; S[u].bad=T[c].bad; S[u].good=T[c].good; S[u++].n=c;}
 qsort(S,u,sizeof(SORTER),sort_compare); c=S[0].n; T[n].child=S[0].n;
 for (i=1;i<u;i++) {T[c].sibling=S[i].n; c=S[i].n;} T[c].sibling=0;}

typedef struct {float rat; uint32 n;} RSORTER;
static int ratio_sort_compare(const void *x, const void *y)
{RSORTER A=*((RSORTER*) x); RSORTER B=*((RSORTER*) y);
 if (A.rat>B.rat) return 1; if (A.rat<B.rat) return -1; return 0;}

static void ratio_sort_childrenA(PN_NODE *T,uint32 node)
{RSORTER S[256]; int i,u=0; uint32 n=T[node].child;
 S[u].rat=T[n].rat; S[u++].n=n;
 while (T[n].sibling) {n=T[n].sibling; S[u].rat=T[n].rat; S[u++].n=n;}
 qsort(S,u,sizeof(RSORTER),ratio_sort_compare); n=S[0].n; T[node].child=S[0].n;
 for (i=1;i<u;i++) {T[n].sibling=S[i].n; n=S[i].n;} T[n].sibling=0;}

static void spit_ratio(float x)
{float l=log(x); printf(" ");
 if (x==FLOAT_INF) printf("+INF  "); else if (x==0.0) printf("-INF  ");
 else if (l>=9.9995) printf("+%.2f",l);
 else if (l<=-9.9995) printf("%.2f",l);
 else if (l>=0.0) printf("+%.3f",l); else printf("%.3f",l); printf("  ");}

static int LIM;
#define GET_BIT(A,x) (A[(x)>>6]&(1ULL<<((x)&0x3f)))
#define SET_BIT(A,x) A[(x)>>6]|=(1ULL<<((x)&0x3f))
static void size_walk(PN_NODE *T,uint32 n,uint64 *A) // trans is bogus?
{if (T[n].loop || T[n].flags&4) n=T[n].trans;
 if (GET_BIT(A,n)) return; LIM++; if (LIM>=1000000) return;
 SET_BIT(A,n); n=T[n].child; if (!n) return;
 size_walk(T,n,A); while ((n=T[n].sibling)) size_walk(T,n,A);}

static uint64 size_of_subtree(PN_NODE *T,uint32 n) // dubious with trans
{uint64 *ARR; uint64 sz,u,ct=0; n=T[n].trans; if (!T[n].child) return 1;
 sz=(T[0].size+127)/64; ARR=malloc(sz*8+16384); LIM=0;
 for (u=0;u<sz;u++) ARR[u]=0; size_walk(T,n,ARR);
 for (u=0;u<sz;u++) ct+=POPCNT(ARR[u]); free(ARR); return ct;}

static void info_book_node(PN_NODE *T,typePOS *POS,uint32 n,boolean b)
{uint32 u,k=T[n].child; char A[8]; boolean known; uint64 sz;
 printf("Val: "); spit_ratio(T[n].rat); spit(T+n);
 printf(" %lld",size_of_subtree(T,n)); if (b) printf(" *"); printf("\n");
 while (k)
 {printf("%s ", Notate(A,T[k].move)); spit_ratio(T[k].rat);
  spit(T+k); u=T[k].loop?T[k].trans:k; sz=size_of_subtree(T,k); known=FALSE;
  if (sz==1)
  {MakeMove(POS,T[k].move); if (known_find(POS->DYN->HASH)!=-1) known=TRUE;
   UnmakeMove(POS,T[k].move,TRUE);}
  printf(" %lld%s %d %d\n",sz,
	 (known || T[k].flags&4 || T[k].flags&64) ? "*" : "",
	 T[u].killer,T[k].size); k=T[k].sibling;}}

static void output_book(PN_NODE *T,typePOS *POS,uint32 n,uint64 Z,boolean b)
{uint32 t=T[n].loop?T[n].trans:n;
 printf ("BOOK %d %d %d %llx\n",n,t,T[n].parent,Z);
 ratio_sort_childrenA(T,t); info_book_node(T,POS,t,b); SortChildren(T,t);}

static uint32 align_booktreeQ(PN_NODE *T,typePOS *POS,uint32 n,boolean *b)
{typeDYNAMIC *D; *b=FALSE; D=POS->DYN_ROOT+1;
 while (D!=POS->DYN && n)
 {D++; n=iTreeMove(T,D->mv,n); if (!T[n].bad || !T[n].good) *b=TRUE;
  if (!n && known_find(D->HASH)!=-1) {follow_known(POS,D); return FALSE;}
  if (T[n].loop || T[n].flags&4) n=T[n].trans;
  if (!T[n].child && known_find(D->HASH)!=-1) return follow_known(POS,D);}
 return n;}

static boolean doit_query(PN_NODE *T,typePOS *POS)
{uint32 t,n; boolean b; n=align_booktreeQ(T,POS,1,&b);
 if (n==0xffffffff) return TRUE; t=T[n].loop?T[n].trans:n; 
 if (!n) {printf("Not in BOOK\n"); return FALSE;}
 else if (!T[t].child) printf("No children %d 0x%llx %d/%d\n",
				 t,POS->DYN->HASH,T[t].good,T[t].bad);
 else output_book(T,POS,n,POS->DYN->HASH,b);
 printf("Book Query Complete\n"); fflush(stdout); return TRUE;}

static void do_extend(PN_NODE *T,typePOS *POS,char *ml);

static void do_getinfo(PN_NODE *T,typePOS *POS,char *A)
{init_position(POS,"startpos"); parse_moves(POS,A);
 if (!doit_query(T,POS)) do_extend(T,POS,A);}

////////////////////////////////////////////////////////////////////////

#define LOSS_TWIST 0x913bb468de01037fULL
#define WIN_TWIST 0xc066034f796185ebULL

static uint32 NewBookHash(HASH *BH,uint32 BHM,uint64 Z,uint32 n)
{uint32 k=Z&BHM; uint64 u; if (!k) k++;
 while ((u=BH[k].key)) {if (u==Z) return BH[k].node; k++; if (k==(BHM+1)) k=1;}
 BH[k].key=Z; BH[k].node=n; return 0;}

static void InstallKnownResult(HASH *BH,uint32 BHM,uint64 Z,uint32 n)
{uint32 k=Z&BHM; uint64 u; if (!k) k++;
 while ((u=BH[k].key)) {if (u==Z) return; k++; if (k==(BHM+1)) k=1;}
 BH[k].key=Z; BH[k].node=n;}

static uint32 IsKnownResult(HASH *BH,uint32 BHM,uint64 Z)
{uint32 k=Z&BHM; uint64 u; if (!k) k++;
 while ((u=BH[k].key)) {if (u==Z) return BH[k].node; k++; if (k==(BHM+1)) k=1;}
 return FALSE;}

static void MakeAllHash(HASH *AH,uint32 AHM,uint32 n,uint64 Z)
{uint32 k=Z&AHM; while (AH[k].key) {k++; if (k==(AHM+1)) k=0;}
 AH[k].key=Z; AH[k].node=n;}

#define FLOAT_INF 9007199254740992.0
#define TEMPO (1.0)
#define TEMPO2 (TEMPO*TEMPO)

static boolean node_ratio(PN_NODE *T,uint32 n)
{float r,x=T[n].rat; uint32 c;
 c=T[n].child; if (!c) return FALSE; r=T[c].rat;
 if (x==0.0 || x==FLOAT_INF) return FALSE;
 while ((c=T[c].sibling)) r=MIN(r,T[c].rat);
 if (r==FLOAT_INF) T[n].rat=0.0;
 else if (r==0.0) T[n].rat=FLOAT_INF; else T[n].rat=TEMPO2/r;
 return (x!=T[n].rat);}

static void RatioBacktrack(PN_NODE *T,uint32 n)
{uint32 k,w; if (!n) return; w=T[n].loop?T[n].trans:n;
 if (node_ratio(T,w)) RatioBacktrack(T,T[w].parent);
 k=T[w].loop; if (!k) return;
 while (k!=w && k) // checking k not really necessary here (but is below)
 {float r=T[k].rat; T[k].rat=T[w].rat;
  if (r!=T[w].rat) RatioBacktrack(T,T[k].parent); k=T[k].loop;}}

static void BacktrackLoop(PN_NODE *T,uint32 node,HASH*,uint32,HASH*,uint32);

static void win_installer
(PN_NODE *T,uint32 n,HASH *BH,uint32 BHM,HASH *AH,uint32 AHM)
{uint32 u,k; uint64 Z,z; Z=T[n].hash; InstallKnownResult(BH,BHM,Z^WIN_TWIST,n);
 T[n].rat=FLOAT_INF; RatioBacktrack(T,T[n].parent);
 T[n].loop=0; T[n].trans=n; T[n].bad=0; T[n].good=MY_INFoo; k=Z&AHM;
 while ((z=AH[k].key))
 {u=AH[k].node; k++; if (k==(AHM+1)) k=0;
  if ((z==Z) && u!=n)
  {if (!T[u].good) continue; T[u].bad=0; T[u].good=MY_INFoo;
   T[u].flags|=4; T[u].trans=n; T[u].loop=0; T[u].child=0;
   BacktrackLoop(T,T[u].parent,BH,BHM,AH,AHM);
   T[u].rat=FLOAT_INF; RatioBacktrack(T,T[u].parent);}}}

static void loss_installer
(PN_NODE *T,uint32 n,HASH *BH,uint32 BHM,HASH *AH,uint32 AHM)
{uint32 u,k; uint64 Z,z; Z=T[n].hash; k=Z&AHM; T[n].good=0; T[n].bad=MY_INFoo;
 InstallKnownResult(BH,BHM,Z^LOSS_TWIST,n); T[n].loop=0; T[n].trans=n;
 T[n].rat=0.0; RatioBacktrack(T,T[n].parent);
 while ((z=AH[k].key))
 {u=AH[k].node; k++; if (k==(AHM+1)) k=0;
  if ((z==Z) && u!=n)
  {if (!T[u].bad) continue; T[u].good=0; T[u].bad=MY_INFoo;
   T[u].flags|=4; T[u].trans=n; T[u].loop=0; T[u].child=0;
   BacktrackLoop(T,T[u].parent,BH,BHM,AH,AHM);
   T[u].rat=0.0; RatioBacktrack(T,T[u].parent);}}}

static boolean UpdateNode
(PN_NODE *T,uint32 n,HASH *BH,uint32 BHM,HASH *AH,uint32 AHM)
{uint32 c,v,w=T[n].good,l=T[n].bad; if (!T[n].child) return TRUE;
 SortChildren(T,n); if (!w || !l) return FALSE;
 c=T[n].child; v=T[c].bad; T[n].bad=T[c].good;
 while (T[c].sibling) {c=T[c].sibling; v+=T[c].bad;if (v>MY_INFoo) v=MY_INFoo;}
 if (w && !v && !(T[n].flags&16) &&
     !IsKnownResult(BH,BHM,T[n].hash^LOSS_TWIST))
   loss_installer(T,n,BH,BHM,AH,AHM);
 if (l && !T[n].bad && (T[n].flags&16) &&
     !IsKnownResult(BH,BHM,T[n].hash^WIN_TWIST))
   win_installer(T,n,BH,BHM,AH,AHM);
 T[n].good=v; if (v==w && l==T[n].bad) return FALSE; return TRUE;}

static void BacktrackLoop
(PN_NODE *T,uint32 node,HASH *BH,uint32 BHM,HASH *AH,uint32 AHM)
{uint32 n,k; if (!node) return; n=T[node].loop?T[node].trans:node;
 if (UpdateNode(T,n,BH,BHM,AH,AHM)) BacktrackLoop(T,T[n].parent,BH,BHM,AH,AHM);
 k=T[n].loop; if (!k) return;
 while (k!=n && k)
 {uint32 g=T[k].good,b=T[k].bad; T[k].good=T[n].good; T[k].bad=T[n].bad;
  if (g!=T[n].good || b!=T[n].bad) BacktrackLoop(T,T[k].parent,BH,BHM,AH,AHM);
  k=T[k].loop;}}

static void iNewNode
(typePOS *POS,uint16 mv,PN_NODE *NEW,uint32 parent, // g/b 1/1 if non-terminal
 uint32 good,uint32 bad,HASH *BH,uint32 BHM,HASH *AH,uint32 AHM)
{uint32 r,t,c,node=NEW[0].size++; uint64 HASH=POS->DYN->HASH; uint16 ML[256];
 if (!(node&1023)) printf("node %d\n",node);
 NEW[node].good=good; NEW[node].bad=bad;
 NEW[node].hash=HASH; MakeAllHash(AH,AHM,node,HASH);
 NEW[node].parent=parent; NEW[node].sibling=NEW[node].child=0;
 NEW[node].move=mv; NEW[node].size=NEW[node].killer=0; NEW[node].size=0;
 NEW[node].flags=POS->wtm?16:0; NEW[node].trans=node; NEW[node].loop=0;
 if (!NEW[parent].child) NEW[parent].child=node; // should parent(1) be 0?
 else {c=NEW[parent].child; while (NEW[c].sibling) c=NEW[c].sibling;
       NEW[c].sibling=node;} // worry patt/rep, TBs/oppB are covered in search
 if (!GenMoves(POS,ML)) // stalemate, combines FICS/International
 {int wc=POPCNT(wBitboardOcc),bc=POPCNT(bBitboardOcc); // does not install
  if (POS->wtm && wc<bc) {NEW[node].bad=0; NEW[node].good=MY_INFoo; return;}
  if (POS->wtm && wc>=bc) {NEW[node].good=0; NEW[node].bad=MY_INFoo; return;}
  if (!POS->wtm) {NEW[node].bad=0; NEW[node].good=MY_INFoo; return;}}
 for (r=4;r<=POS->DYN->rev;r+=2) if (((POS->DYN-r)->HASH)==HASH)
 {if (POS->wtm) {NEW[node].good=0; NEW[node].bad=MY_INFoo; return;}
  if (!POS->wtm) {NEW[node].bad=0; NEW[node].good=MY_INFoo; return;}}
 if (POS->wtm && (t=IsKnownResult(BH,BHM,HASH^WIN_TWIST)))
 {NEW[node].good=MY_INFoo; NEW[node].bad=0;
  NEW[node].flags|=4; NEW[node].trans=t; NEW[node].loop=0; return;}
 if (!POS->wtm && (t=IsKnownResult(BH,BHM,HASH^LOSS_TWIST)))
 {NEW[node].bad=MY_INFoo; NEW[node].good=0;
  NEW[node].flags|=4; NEW[node].trans=t; NEW[node].loop=0; return;}
 if (POS->wtm && !NEW[node].bad) win_installer(NEW,node,BH,BHM,AH,AHM);
 if (!POS->wtm && !NEW[node].good) loss_installer(NEW,node,BH,BHM,AH,AHM);
 if (POS->DYN->rev==0 && (t=NewBookHash(BH,BHM,HASH,node)))
 {NEW[node].trans=t; NEW[node].good=NEW[t].good; NEW[node].bad=NEW[t].bad;
  NEW[node].loop=NEW[t].loop?NEW[t].loop:t; NEW[t].loop=node;}}

static void NewNode // use TEMPO here?
(typePOS *POS,uint16 mv,PN_NODE *NEW,uint32 parent, // g/b 1/1 if non-terminal
 uint32 good,uint32 bad,HASH *BH,uint32 BHM,HASH *AH,uint32 AHM)
{uint32 g,b,w=NEW[0].size; iNewNode(POS,mv,NEW,parent,good,bad,BH,BHM,AH,AHM);
 if (NEW[w].flags&4 || NEW[w].loop)
 {uint32 t=NEW[w].trans; if (t!=w) {NEW[w].rat=NEW[t].rat; return;}}
 g=NEW[w].good; b=NEW[w].bad; if (b && g) NEW[w].rat=((float) g)/((float) b);
 if (b==0) NEW[w].rat=FLOAT_INF; if (g==0) NEW[w].rat=0.0;}

static double TB5RAT=1048576.0;
static void recurse(typePOS *POS,PN_NODE *OLD,uint32 n,PN_NODE *NEW,uint32 par,
		    HASH *BH,uint32 BHM,HASH *AH,uint32 AHM)
{uint32 cur,c,good,bad,t; uint16 mv=OLD[n].move;
 if (OLD[n].flags&4 || OLD[n].loop) t=OLD[n].trans; else t=n;
 SortChildren(OLD,t); c=OLD[t].child;
 if (!c) {good=OLD[t].good; bad=OLD[t].bad;} else {good=1; bad=1;}
 cur=NEW[0].size; NewNode(POS,mv,NEW,par,good,bad,BH,BHM,AH,AHM);
 NEW[cur].flags|=OLD[t].flags&1; // copy TB flag info
 if (NEW[cur].flags&1 && !c && NEW[cur].bad && NEW[cur].good)
   NEW[cur].rat*=(NEW[cur].flags&16)?TB5RAT:1/TB5RAT; // adjust ratio
 if (!c) return; if (NEW[cur].flags&4 || NEW[cur].loop) cur=NEW[cur].trans;
 if (!NEW[cur].bad || !NEW[cur].good || NEW[cur].child) return;
 do {MakeMoveCheck(POS,OLD[c].move,FALSE);
     recurse(POS,OLD,c,NEW,cur,BH,BHM,AH,AHM); // cur is parent
     UnmakeMove(POS,OLD[c].move,TRUE);} while ((c=OLD[c].sibling));
 BacktrackLoop(NEW,cur,BH,BHM,AH,AHM); RatioBacktrack(NEW,cur);}

////////////////////////////////////////////////////////////////////////

static void mknode(typePOS *POS,PN_NODE *OLD,uint32 o,PN_NODE *NEW,uint32 par,
		   HASH *BH,uint32 BHM,HASH *AH,uint32 AHM)
{uint32 good,bad; uint32 cur;
 if (!OLD[o].child) {good=OLD[o].good; bad=OLD[o].bad;} else {good=1; bad=1;}
 MakeMoveCheck(POS,OLD[o].move,FALSE);
 cur=NEW[0].size; NewNode(POS,OLD[o].move,NEW,par,good,bad,BH,BHM,AH,AHM);
 UnmakeMove(POS,OLD[o].move,TRUE);
 NEW[cur].flags|=OLD[o].flags&1; // copy TB flag info, adjust ratio
 if (NEW[cur].flags&1 && !OLD[o].child && NEW[cur].bad && NEW[cur].good)
   NEW[cur].rat*=(NEW[cur].flags&16)?TB5RAT:1/TB5RAT;}

static boolean insert_path(typePOS *POS,PN_NODE *OLD,PN_NODE *NEW,
			   HASH *BH,uint32 BHM,HASH *AH,uint32 AHM)
{uint64 HASHL[1024]; uint32 NEWnode[1024]; uint32 u,e,c,nptr=1,cn=1,co=1;
 typeDYNAMIC *D=POS->DYN_ROOT+1; HASHL[0]=D->HASH; NEWnode[0]=1;
 // printf("INSERT PATH\n"); dump_path(POS);
 while (TRUE) // unclear: if merging into NEW maybe this is not correct?
 {LOOP: D++; if (D>POS->DYN) return NEW[cn].child; // already expanded in NEW
  for (u=0;u<nptr;u++) if (HASHL[u]==D->HASH)
  {cn=NEWnode[u]; nptr=u; co=TreeMove(OLD,D->mv,co); goto LOOP;}
  e=TreeMove(NEW,D->mv,cn); if (!e) break;
  cn=e; HASHL[nptr]=D->HASH; NEWnode[nptr++]=cn; co=TreeMove(OLD,D->mv,co);}
 // situation: co is parent in OLD tree, cn is parent in NEW tree
 if (D!=POS->DYN) // hack when NEW is not on frontier :( // should NOT happen
 {uint16 mv=POS->DYN->mv; printf("%d %d\n",D,POS->DYN); dump_path(POS);
  raise(SIGTRAP); UnmakeMove(POS,mv,TRUE);
  insert_path(POS,OLD,NEW,BH,BHM,AH,AHM); MakeMoveCheck(POS,mv,FALSE);
  insert_path(POS,OLD,NEW,BH,BHM,AH,AHM); return TRUE;}
 uint16 mv=POS->DYN->mv; UnmakeMove(POS,mv,TRUE);
 SortChildren(OLD,co); c=OLD[co].child;
 do {mknode(POS,OLD,c,NEW,cn,BH,BHM,AH,AHM); if (!OLD[c].good) break;}
 while ((c=OLD[c].sibling)); MakeMoveCheck(POS,mv,FALSE);
 BacktrackLoop(NEW,cn,BH,BHM,AH,AHM); RatioBacktrack(NEW,cn); return FALSE;}

static void new_recurse(typePOS *POS,PN_NODE *OLD,uint32 n,PN_NODE *NEW,
			HASH *BH,uint32 BHM,HASH *AH,uint32 AHM)
{uint32 c,t; if (OLD[n].flags&4 || OLD[n].loop) t=OLD[n].trans; else t=n;
 if (insert_path(POS,OLD,NEW,BH,BHM,AH,AHM)) return;
 SortChildren(OLD,t); c=OLD[t].child; if (!c) return;
 do {MakeMoveCheck(POS,OLD[c].move,FALSE);
     new_recurse(POS,OLD,c,NEW,BH,BHM,AH,AHM);
     UnmakeMove(POS,OLD[c].move,TRUE); if (!OLD[c].good) break;} // prune
 while ((c=OLD[c].sibling));}

////////////////////////////////////////////////////////////////////////

static void do_savebook(PN_NODE *T,char* fn);
static PN_NODE *ROOT_TREE=NULL; static typePOS RPOS[1];
static HASH *BOOK_HASH=NULL; static uint32 BOOK_HASH_MASK; // static bad
static HASH *ALL_HASH=NULL; static uint32 ALL_HASH_MASK; // for debugger

////////////////////////////////////////////////////////////////////////

static void prune_recurse(typePOS *POS,PN_NODE *OLD,uint32 n,PN_NODE *NEW,
			  uint32 w,HASH *BH,uint32 BHM,HASH *AH,uint32 AHM)
{uint32 r,c,good,bad,t; uint16 mv=OLD[n].move;
 if (OLD[n].flags&4 || OLD[n].loop) t=OLD[n].trans; else t=n; c=OLD[t].child;
 if (!c) {good=OLD[t].good; bad=OLD[t].bad;} else {good=1; bad=1;}
 r=NEW[0].size; NewNode(POS,mv,NEW,w,good,bad,BH,BHM,AH,AHM);
 NEW[r].flags|=OLD[t].flags&1; // copy TB flag info
 if (NEW[r].flags&1 && !c && NEW[r].bad && NEW[r].good)
   NEW[r].rat*=(NEW[r].flags&16)?TB5RAT:1/TB5RAT; // adjust ratio
 if (!c) return; if (NEW[r].flags&4 || NEW[r].loop) r=NEW[r].trans;
 if (!NEW[r].bad || !NEW[r].good || NEW[r].child) return; SortChildren(OLD,c);
 do {MakeMoveCheck(POS,OLD[c].move,FALSE);
     prune_recurse(POS,OLD,c,NEW,r,BH,BHM,AH,AHM);
     UnmakeMove(POS,OLD[c].move,TRUE); if (!OLD[c].good) break;}
 while ((c=OLD[c].sibling));
 BacktrackLoop(NEW,r,BH,BHM,AH,AHM); RatioBacktrack(NEW,r);}

static void do_prune(PN_NODE *T,typePOS *POS)
{PN_NODE *NEW; HASH *BH,*AH; uint32 BHM,AHM,sz,all,u;
 sz=T[0].size+65536; NEW=malloc(sz*sizeof(PN_NODE));
 all=2<<BSR(sz); BH=malloc(all*sizeof(HASH));
 for (u=0;u<all;u++) BH[u].key=0; BHM=all-1;
 all=2<<BSR(sz+sz/2); AH=malloc(all*sizeof(HASH));
 for (u=0;u<all;u++) AH[u].key=0; AHM=all-1;
 NEW[0].size=1; NEW[0].child=NEW[0].parent=NEW[0].sibling=0; NEW[0].hash=0;
 NEW[0].good=NEW[0].bad=0; NEW[0].loop=NEW[0].trans=0;
 NEW[0].move=NEW[0].flags=NEW[0].killer=0; init_position(POS,"startpos");
 prune_recurse(POS,T,1,NEW,0,BH,BHM,AH,AHM);
 do_savebook(NEW,"BOOKS/PRUNED.book");
 free(BOOK_HASH); free(ALL_HASH); BOOK_HASH=BH; ALL_HASH=AH;
 BOOK_HASH_MASK=BHM; ALL_HASH_MASK=AHM; free(ROOT_TREE); ROOT_TREE=NEW;
 do_getinfo(ROOT_TREE,RPOS,"");}

////////////////////////////////////////////////////////////////////////

static uint32 align_booktreeB(PN_NODE *T,typePOS *POS,uint32 n)
{typeDYNAMIC *D; D=POS->DYN_ROOT+1;
 while (D!=POS->DYN && n) {D++; n=TreeMove(T,D->mv,n);} return n;}

static void do_directline(PN_NODE *T,typePOS *POS,char *ml,boolean SAVE);
static void doit_merge(PN_NODE *T1,char *fn);

static void do_extend(PN_NODE *T,typePOS *POS,char *ml)
{uint16 mv; uint32 n; init_position(POS,"startpos"); parse_moves(POS,ml);
 mv=POS->DYN->mv; UnmakeMove(POS,mv,TRUE); n=align_booktreeB(T,POS,1);
 if (!n) {printf("Failure to align\nBook Query Complete\n"); return;}
 if (!T[n].good || !T[n].bad) // merge with DIRECTLINE
 {PN_NODE *T2; T2=malloc(0x10000); do_directline(T2,POS,ml,TRUE); free(T2);
  doit_merge(T,"BOOKS/DIRECTLINE.book"); do_getinfo(ROOT_TREE,POS,ml); return;}
 MakeMoveCheck(POS,mv,FALSE);
 NewNode(POS,mv,T,n,1,1,BOOK_HASH,BOOK_HASH_MASK,ALL_HASH,ALL_HASH_MASK);
 BacktrackLoop(T,n,BOOK_HASH,BOOK_HASH_MASK,ALL_HASH,ALL_HASH_MASK);
 RatioBacktrack(T,n); do_getinfo(T,POS,ml);}

////////////////////////////////////////////////////////////////////////

static void hack_wintree(typePOS *POS,PN_NODE *tree)
{tWIN WT[1]; FILE *F; int u; char STR[256]; STR[0]=0;
 WT->WIN=malloc(sizeof(WIN_NODE)*(1<<24));
 WT->TRANS=malloc(sizeof(tTRANS)*(1<<24)); WT->MASK=(1<<24)-1;
 WT->LIST=malloc(sizeof(uint32)*(1<<24));
 for (u=0;u<(1<<24);u++) WT->TRANS[u].key=0;
 for (u=0;u<(1<<24);u++) WT->WIN[u].data=0; do_wintree(POS,tree,1,WT);
 strcat(STR,"PROOFS/TEMP.proof"); F=fopen(STR,"w");
 printf("Writing to %s (%d nodes)\n",STR,WT->WIN[0].data&0x3fffffff);
 fwrite(&(WT->WIN[0].data),4,1,F); fwrite(&(WT->WIN[0].move),2,1,F);
 for (u=0;u<(WT->WIN[0].move);u++) fwrite(&(WT->ML[u]),2,1,F);
 for (u=1;u<(WT->WIN[0].data&0x3fffffff);u++)
 {fwrite(&(WT->WIN[u].data),4,1,F); WT->WIN[u].data=0;
  fwrite(&(WT->WIN[u].move),2,1,F);}
 printf("Proof file completed\n");
 fclose(F); free(WT->WIN); free(WT->TRANS); free(WT->LIST);}

extern int NUM_TBS;
static void do_search(typePOS *POS,char *ml) // ml has nodelim token
{uint32 nodelim; char nodestr[16]; PN_NODE *tree;
 init_position(POS,"startpos"); sscanf(ml,"%s",nodestr);
 ml+=strlen(nodestr)+1; nodelim=atoi(nodestr); parse_moves(POS,ml);
 tree=MALLOC((2*MAX_PLY+nodelim),sizeof(PN_NODE)); // HASH_MASK is excessive
 HASH_MASK=(2<<BSR(nodelim))-1; if (POS->ZTAB) free(POS->ZTAB);
 POS->ZTAB=CALLOC((HASH_MASK+1),sizeof(HASH)); if (!NUM_TBS) load_tbs("FLAT");
 TIME_LIMIT=10000.0; pn_search(tree,POS,nodelim,FALSE); // outputs result
 if (!tree[1].bad || !tree[1].good) hack_wintree(POS,tree); free(tree);}

static void do_goodbad(PN_NODE *T,typePOS *POS,char *ml)
{uint32 k,p,n,g,b; char G[16],B[16];
 init_position(POS,"startpos"); sscanf(ml,"%s %s",G,B);
 ml+=strlen(G)+1+strlen(B)+1; g=atoi(G); b=atoi(B);
 parse_moves(POS,ml); n=align_booktreeB(T,POS,1);
 if (!n || !b || !g || T[n].child || !T[n].good || !T[n].bad) return;
 T[n].good=g; T[n].bad=b; T[n].rat=((float) g)/((float) b); p=T[n].parent;
 BacktrackLoop(T,p,BOOK_HASH,BOOK_HASH_MASK,ALL_HASH,ALL_HASH_MASK);
 RatioBacktrack(T,p); k=T[n].loop; if (!k) return;
 while (k!=n && k)
 {T[k].good=g; T[k].bad=b; T[k].rat=T[n].rat; p=T[k].parent;
  BacktrackLoop(T,p,BOOK_HASH,BOOK_HASH_MASK,ALL_HASH,ALL_HASH_MASK);
  RatioBacktrack(T,p); k=T[k].loop;}}

////////////////////////////////////////////////////////////////////////

static PN_NODE* do_subtree
(PN_NODE *T,typePOS *POS,char *ml,char *fn,boolean GI,boolean OW)
{PN_NODE *NEW; HASH *BH,*AH; uint32 BHM,AHM,sz,all,u,n,r; typeDYNAMIC *D;
 init_position(POS,"startpos"); parse_moves(POS,ml);
 n=align_booktreeB(T,POS,1); if (!n) return NULL;
 sz=T[0].size+65536; NEW=malloc(sz*sizeof(PN_NODE));
 all=2<<BSR(sz); BH=malloc(all*sizeof(HASH));
 for (u=0;u<all;u++) BH[u].key=0; BHM=all-1;
 all=2<<BSR(sz+sz/2); AH=malloc(all*sizeof(HASH));
 for (u=0;u<all;u++) AH[u].key=0; AHM=all-1;
 NEW[0].child=NEW[0].parent=NEW[0].sibling=0; NEW[0].hash=0;
 NEW[0].good=NEW[0].bad=0; NEW[0].loop=NEW[0].trans=0;
 NEW[0].move=NEW[0].flags=NEW[0].killer=0; if (n==1) {r=0; goto LEAP;}
 NEW[1].child=NEW[1].parent=NEW[1].sibling=0; NEW[1].rat=1.0;
 NEW[1].good=NEW[1].bad=1; NEW[1].loop=NEW[1].size=0;
 NEW[1].trans=1; NEW[1].move=NEW[1].killer=0; NEW[1].flags=16;
 r=1; D=POS->DYN_ROOT+1; NEW[1].hash=D->HASH; // problem when ml is ""
 while (D!=(POS->DYN)) // NewNode has HASH difficulty
 {D++; r++; NEW[r].move=D->mv; NEW[r].parent=r-1; NEW[r-1].child=r;
  NEW[r].good=NEW[r].bad=1; NEW[r].flags=(r&1)?16:0; NEW[r].rat=1.0;
  NEW[r].child=NEW[r].sibling=0; NEW[r].hash=D->HASH;
  NEW[r].loop=NEW[r].size=0; NEW[r].trans=r; NEW[r].killer=0;
  MakeAllHash(AH,AHM,r,D->HASH);} LEAP: NEW[0].size=r+1;
 new_recurse(POS,T,n,NEW,BH,BHM,AH,AHM); if (fn) do_savebook(NEW,fn);
 // recurse(POS,T,n,NEW,r,BH,BHM,AH,AHM); if (fn) do_savebook(NEW,fn);
 if (!OW) {free(T); return NEW;} // otherwise input T is ROOT_TREE
 free(BOOK_HASH); free(ALL_HASH); BOOK_HASH=BH; ALL_HASH=AH;
 BOOK_HASH_MASK=BHM; ALL_HASH_MASK=AHM; free(ROOT_TREE); ROOT_TREE=NEW;
 if (GI) do_getinfo(ROOT_TREE,RPOS,ml); return NULL;}

////////////////////////////////////////////////////////////////////////

// static double TB5RAT=1048576.0; // now defined above
#define TEMPO (1.0)
#define TEMPO2 (TEMPO*TEMPO)

static void initial_book_scan
(PN_NODE *T,HASH *BH,uint32 BHM,HASH *AH,uint32 AHM)
{uint32 u; for (u=1;u<T[0].size;u++)
 {MakeAllHash(AH,AHM,u,T[u].hash);
  if (T[u].flags&4 || T[u].flags&64) continue;
  if (T[u].flags&1 && T[u].rat!=-1 && T[u].bad && T[u].good)
    T[u].rat*=(T[u].flags&16)?TB5RAT:1/TB5RAT;
  if (!T[u].good && !(T[u].flags&16) &&
      !IsKnownResult(BH,BHM,T[u].hash^LOSS_TWIST))
    InstallKnownResult(BH,BHM,T[u].hash^LOSS_TWIST,u);
  if (!T[u].bad && (T[u].flags&16) &&
      !IsKnownResult(BH,BHM,T[u].hash^WIN_TWIST))
    InstallKnownResult(BH,BHM,T[u].hash^WIN_TWIST,u);}}

static void init_all_ratios(PN_NODE *T)
{uint32 q,i,t,k,c,u,v=0,z=0; float r; uint32 *ARR;
 ARR=malloc(T[0].size*sizeof(uint32));
 for (i=1;i<T[0].size;i++)
 {k=T[i].loop; t=k?T[i].trans:i; if (!k || T[t].child) continue;
  k=T[t].loop; while (k!=t) {T[k].rat=T[t].rat; k=T[k].loop;}}
 for (i=1;i<T[0].size;i++)
  if ((!T[i].loop || T[i].trans==i) && T[i].rat==-1) ARR[v++]=i;
 while (T[1].rat==-1 && z<1000)
 {u=v; if (0) printf("PASS %d: left %d\n",++z,u); v=0;
  for (i=0;i<u;i++)
  {q=ARR[i]; t=T[q].loop?T[q].trans:q;
   if (T[t].rat!=-1.0) {ARR[v++]=ARR[i]; continue;}
   c=T[t].child; if (!c || T[c].rat==-1.0) {ARR[v++]=ARR[i]; continue;}
   k=T[t].loop; r=FLOAT_INF; r=MIN(r,T[c].rat);
   while ((c=T[c].sibling)) {if (T[c].rat==-1.0) break; r=MIN(r,T[c].rat);}
   if (c) {ARR[v++]=ARR[i]; continue;} if (r==FLOAT_INF) T[t].rat=0.0;
   else if (r==0.0) T[t].rat=FLOAT_INF; else T[t].rat=TEMPO2/r;
   if (k) while (k!=t) {T[k].rat=T[t].rat; k=T[k].loop;}}} free(ARR);}

static void InitBookHash(PN_NODE *T,typePOS *POS,HASH *BH,uint32 BHM,uint32 n)
{uint32 c; if (T[n].flags&4) return;
 if (T[n].hash!=POS->DYN->HASH)
  printf("HASH prob %d %llx %llx\n",n,T[n].hash,POS->DYN->HASH);
 if (T[n].loop && T[n].trans!=n) return;
 if (POS->DYN->rev==0) NewBookHash(BH,BHM,POS->DYN->HASH,n); // could be there!
 c=T[n].child; if (!c) return;
 if (!MakeMoveCheck(POS,T[c].move,FALSE)) return; // hmm
 InitBookHash(T,POS,BH,BHM,c); UnmakeMove(POS,T[c].move,TRUE);
 while ((c=T[c].sibling))
 {if (!MakeMoveCheck(POS,T[c].move,FALSE)) continue;
  InitBookHash(T,POS,BH,BHM,c); UnmakeMove(POS,T[c].move,TRUE);}}
// should load all won hashes, not just those reachable from root!

////////////////////////////////////////////////////////////////////////

static PN_NODE* new_book(typePOS *POS,char *fn)
{PN_NODE *T; T=malloc(65536*sizeof(PN_NODE)); printf("New book created\n");
 T[0].child=T[0].parent=T[0].sibling=0; T[0].good=T[0].bad=0;
 T[0].loop=T[0].trans=0; T[0].hash=0;
 T[0].move=T[0].flags=T[0].killer=0; // strcpy(BOOK_NAME,fn); // fn is a hack
 T[0].size=2; T[1].good=T[1].bad=1; T[1].hash=POS->DYN->HASH; T[1].rat=1.0;
 T[1].parent=T[1].child=T[1].sibling=0; T[1].trans=1; T[1].loop=0;
 T[1].size=0; T[1].move=0; T[1].flags=16; T[1].killer=0; uint32 u,all=65536;
 BOOK_HASH=CALLOC(all,sizeof(HASH)); for (u=0;u<all;u++) BOOK_HASH[u].key=0;
 ALL_HASH=CALLOC(all,sizeof(HASH)); for (u=0;u<all;u++) ALL_HASH[u].key=0;
 BOOK_HASH_MASK=all-1; ALL_HASH_MASK=all-1;
 MakeAllHash(ALL_HASH,ALL_HASH_MASK,1,T[1].hash); return T;}

static void LoadNode(FILE *F,PN_NODE *P,uint32 w)
{fread(&(P->good),4,1,F); fread(&(P->bad),4,1,F); fread(&(P->size),4,1,F);
 fread(&P->hash,8,1,F); fread(&(P->trans),4,1,F); fread(&(P->loop),4,1,F);
 fread(&(P->parent),4,1,F); fread(&(P->sibling),4,1,F);
 fread(&(P->child),4,1,F); fread(&(P->move),2,1,F);
 fread(&(P->killer),1,1,F); fread(&(P->flags),1,1,F); P->killer=0; // ok
 if (P->child || (P->loop && P->trans!=w)) P->rat=-1.0;
 else {if (!(P->bad)) P->rat=FLOAT_INF; else P->rat=(TEMPO*P->good)/P->bad;}}

static PN_NODE* load_book(typePOS *POS,char *fn) // on startup
{char A[256]; uint32 u,sz,all,osz; FILE *F; PN_NODE *T; strcpy(A,fn);
 F=fopen(A,"r"); if (!F) {strcpy(A,"BOOKS/"); strcat(A,fn); F=fopen(A,"r");}
 if (!F) {printf("File %s not found\n",A); return NULL;}
 T=malloc(sizeof(PN_NODE)); LoadNode(F,T,0); sz=T[0].size;
 printf("BOOK %s has %d positions\n",fn,sz);
 osz=sz+65536+sz/8; T=realloc(T,osz*sizeof(PN_NODE));
 for (u=1;u<sz;u++) LoadNode(F,T+u,u); fclose(F);
 all=2<<BSR(osz); BOOK_HASH_MASK=all-1; BOOK_HASH=CALLOC(all,sizeof(HASH));
 for (u=0;u<all;u++) BOOK_HASH[u].key=0;
 all=2<<BSR(osz+osz/2); ALL_HASH_MASK=all-1; ALL_HASH=CALLOC(all,sizeof(HASH));
 for (u=0;u<all;u++) ALL_HASH[u].key=0;
 InitBookHash(T,POS,BOOK_HASH,BOOK_HASH_MASK,1);
 initial_book_scan(T,BOOK_HASH,BOOK_HASH_MASK,ALL_HASH,ALL_HASH_MASK);
 init_all_ratios(T); return T;}

static void do_loadbook(typePOS *POS,char *fn) // other times
{char A[256]; uint32 u,sz,all,osz; FILE *F; PN_NODE *T; strcpy(A,fn);
 F=fopen(A,"r"); if (!F) {strcpy(A,"BOOKS/"); strcat(A,fn); F=fopen(A,"r");}
 if (!F) {printf("File %s not found\n",A); return;}
 if (ROOT_TREE) free(ROOT_TREE); ROOT_TREE=NULL; init_position(POS,"startpos");
 T=malloc(sizeof(PN_NODE)); LoadNode(F,T,0); sz=T[0].size;
 printf("BOOK %s has %d positions\n",fn,sz);
 osz=sz+65536+sz/8; T=realloc(T,osz*sizeof(PN_NODE));
 for (u=1;u<sz;u++) LoadNode(F,T+u,u); fclose(F); ROOT_TREE=T;
 if (BOOK_HASH) free(BOOK_HASH); if (ALL_HASH) free(ALL_HASH);
 all=2<<BSR(osz); BOOK_HASH=CALLOC(all,sizeof(HASH));
 BOOK_HASH_MASK=all-1; for (u=0;u<all;u++) BOOK_HASH[u].key=0;
 all=2<<BSR(osz+osz/2); ALL_HASH=CALLOC(all,sizeof(HASH));
 ALL_HASH_MASK=all-1; for (u=0;u<all;u++) ALL_HASH[u].key=0;
 InitBookHash(T,POS,BOOK_HASH,BOOK_HASH_MASK,1);
 initial_book_scan(T,BOOK_HASH,BOOK_HASH_MASK,ALL_HASH,ALL_HASH_MASK);
 init_all_ratios(T); do_getinfo(T,POS,"");}

static void SaveNode(FILE *F,PN_NODE *P)
{fwrite(&(P->good),4,1,F); fwrite(&(P->bad),4,1,F); fwrite(&(P->size),4,1,F);
 fwrite(&(P->hash),8,1,F); fwrite(&(P->trans),4,1,F); fwrite(&(P->loop),4,1,F);
 fwrite(&(P->parent),4,1,F); fwrite(&(P->sibling),4,1,F);
 fwrite(&(P->child),4,1,F); fwrite(&(P->move),2,1,F);
 fwrite(&(P->killer),1,1,F); fwrite(&(P->flags),1,1,F);}

static void do_savebook(PN_NODE *T,char *fn)
{FILE *F; uint32 u;
 F=fopen(fn,"w"); if (!F) {printf("Failure to open %s\n",fn); return;}
 for (u=0;u<T[0].size;u++) SaveNode(F,T+u);
 printf("BOOK SAVE Wrote to %s, %d positions\n",fn,T[0].size); fclose(F);}

////////////////////////////////////////////////////////////////////////

static char SETUP[4096];
static void do_excludesetup(PN_NODE *T,typePOS *POS,char *ml)
{init_position(POS,"startpos"); parse_moves(POS,ml); strcpy(SETUP,ml);}

static void do_exclude(PN_NODE *T,typePOS *POS,char *ml,boolean SUBEXCLUDE)
{uint32 e,c,n,exc=0,l=0; uint16 EXCLUDE[256]; boolean CHILD,SKIP; char S[16];
 n=align_booktreeB(T,POS,1); if (!n) return; c=T[n].child; CHILD=FALSE;
 while (ml[0])
 {sscanf(ml,"%s",S); EXCLUDE[exc++]=get_move(POS,S);
  ml+=strlen(S); while (ml[0]==' ') ml++;}
 do {SKIP=FALSE; for (e=0;e<exc;e++) if (T[c].move==EXCLUDE[e]) SKIP=TRUE;
     if (!SKIP) {if (!CHILD) T[n].child=c; else T[l].sibling=c; CHILD=1; l=c;}}
 while ((c=T[c].sibling)); T[l].sibling=0;
 if (SUBEXCLUDE) do_subtree(T,POS,SETUP,"BOOKS/SUBEXCLUDE.book",FALSE,TRUE);
 else do_subtree(T,POS,"","BOOKS/EXCLUDED.book",FALSE,TRUE);}

WIN_STRUCT* try_wintree(PN_NODE *T,typePOS *POS,char *ml,uint16 *SKIP);
void expand_and_rehash(WIN_STRUCT *WS,char *fout,boolean SAVE);
void wintree_direct(typePOS *POS,char *fn); // util_wintree.c
void tree_build(typePOS *POS,PN_NODE *OLD,char *ml); // util_treebuild.c

static void direct_portion(PN_NODE *T,typePOS *POS)
{int w=2; typeDYNAMIC *D=POS->DYN_ROOT+1; T[1].flags=16;
 T[1].good=1; T[1].bad=1; T[1].trans=1; T[1].loop=0; T[1].size=0;
 T[1].hash=(D)->HASH; T[1].move=0; T[1].child=2; T[1].sibling=0;
 while (D!=POS->DYN) // does not have BOOK_HASH or ALL_HASH
 {T[w-1].child=w; T[w].parent=w-1; T[w].child=0; T[w].sibling=0;
  T[w].good=1; T[w].bad=1; T[w].size=0; T[w].trans=w; T[w].loop=0;
  T[w].hash=(D+1)->HASH; T[w].flags=(w&1)?16:0; T[w].move=(D+1)->mv; D++; w++;}
 T[0].size=w;}

static void do_directline(PN_NODE *T,typePOS *POS,char *ml,boolean SAVE)
{init_position(POS,"startpos"); parse_moves(POS,ml);
 direct_portion(T,POS); if (SAVE) do_savebook(T,"BOOKS/DIRECTLINE.book");}

static void doit_wintree(PN_NODE *T,typePOS *POS,char *ml)
{uint16 SK[256]; int e,sk=0; uint32 n,c,l=0; uint32 AHM,BHM; HASH *AH,*BH;
 boolean CHILD,SKIP; PN_NODE *U; uint64 sz,sz2; WIN_STRUCT *WS;
 char fn[64]; strcpy(fn,"PROOFS/EXPANDED.proof");
 init_position(POS,"startpos"); parse_moves(POS,ml);
 n=align_booktreeB(T,POS,1); if (!n) {printf("Failed to align\n"); return;}
 if (POS->wtm && T[n].bad) {printf("White-to-move, no wins seen\n"); return;}
 if (known_find(POS->DYN->HASH)!=-1) {wintree_direct(POS,fn); return;}
 c=T[n].child; if (!c) {wintree_direct(POS,fn); return;}
 sz=T[0].size; sz2=sz*sizeof(PN_NODE); U=malloc(sz2); memcpy(U,T,sz2);
 do {if (!POS->wtm && U[c].bad) SK[sk++]=U[c].move;} while ((c=U[c].sibling));
 c=U[n].child; CHILD=FALSE;
 do {SKIP=FALSE; for (e=0;e<sk;e++) if (U[c].move==SK[e]) SKIP=TRUE;
     if (!SKIP) {if (!CHILD) U[n].child=c; else U[l].sibling=c; CHILD=1; l=c;}
     else U[c].parent=0;} while ((c=U[c].sibling));
 U[l].sibling=0; if (!CHILD) {printf("No won children\n"); return;}
#if 0
 BHM=BOOK_HASH_MASK; AHM=ALL_HASH_MASK;
 sz2=(BHM+1)*sizeof(HASH); BH=malloc(sz2); memcpy(BH,BOOK_HASH,sz2);
 sz2=(AHM+1)*sizeof(HASH); AH=malloc(sz2); memcpy(AH,ALL_HASH,sz2);
 BacktrackLoop(U,n,BH,BHM,AH,AHM); SK[sk]=0;
#else // maybe could prune here... much safer than subtree
 BHM=BOOK_HASH_MASK; sz2=(BHM+1)*sizeof(HASH); BH=malloc(sz2);
 AHM=ALL_HASH_MASK; sz2=(AHM+1)*sizeof(HASH); AH=malloc(sz2);
 int u; for (u=0;u<BHM+1;u++) BH[u].key=0; for (u=0;u<AHM+1;u++) AH[u].key=0;
 PN_NODE *V; sz2=sz*sizeof(PN_NODE); V=malloc(sz2);
 uint16 mv=POS->DYN->mv; UnmakeMove(POS,mv,TRUE);
 direct_portion(V,POS); MakeMove(POS,mv);
 printf("Pruning down to solved subtree\n"); SK[sk]=0;
 prune_recurse(POS,U,n,V,V[0].size-1,BH,BHM,AH,AHM); free(U); U=V;
 // new_recurse(POS,U,n,V,BH,BHM,AH,AHM); free(U); U=V; // use new_recurse()
 printf("Done pruning, size is %d\n",U[0].size); fflush(stdout);
#endif
 WS=try_wintree(U,POS,ml,SK); free(U); free(AH); free(BH); fflush(stdout);
 if (WS) // when solved, then expand and rehash, else tree_build() if possible
 {expand_and_rehash(WS,"PROOFS/EXPANDED.proof",TRUE);
  free(WS->W); free(WS->MOVE_LIST); free(WS); return;}
 if (sk) printf("Cannot make SIMPLIFIED.book with unsolved Black moves\n");
 else tree_build(POS,T,ml);}

PN_NODE* try_merge
(PN_NODE *T1,PN_NODE *T2,HASH* *BH,uint32* BHM,HASH* *AH,uint32* AHM);

static void doit_merge(PN_NODE *T1,char *fn)
{char A[256]; uint32 u,sz; FILE *F; PN_NODE *T2,*TC,*MT;
 HASH *AH,*BH; uint32 AHM,BHM; strcpy(A,fn);
 F=fopen(A,"r"); if (!F) {strcpy(A,"BOOKS/"); strcat(A,fn); F=fopen(A,"r");}
 if (!F) {printf("File %s not found\n",A); return;}
 T2=malloc(sizeof(PN_NODE)); LoadNode(F,T2,0); sz=T2[0].size;
 printf("BOOK %s has %d positions\n",fn,sz);
 T2=realloc(T2,(sz+0x10000)*sizeof(PN_NODE));
 for (u=1;u<sz;u++) LoadNode(F,T2+u,u); fclose(F);
 sz=T1[0].size*sizeof(PN_NODE); TC=malloc(sz); memcpy(TC,T1,sz);
 MT=try_merge(TC,T2,&BH,&BHM,&AH,&AHM); free(TC); free(T2); if (!MT) return;
 free(T1); free(BOOK_HASH); free(ALL_HASH); ROOT_TREE=MT;
 BOOK_HASH=BH; BOOK_HASH_MASK=BHM; ALL_HASH=AH; ALL_HASH_MASK=AHM;
 do_savebook(MT,"BOOKS/MERGE.book");}

static void init_stuff(int c,char **I)
{magic_mult_init(); RPOS->DYN_ROOT=CALLOC(MAX_PLY,sizeof(typeDYNAMIC));
 RPOS->DYN=RPOS->DYN_ROOT+1; init_position(RPOS,"startpos");
 RPOS->ZTAB=NULL; init_tb_stuff();
 printf("LOSING CHESS\n"); fflush(stdout); init_known(FALSE);
 if (c>1) ROOT_TREE=load_book(RPOS,I[1]);
 else ROOT_TREE=new_book(RPOS,"BOOKS/NEW.book"); if (ROOT_TREE==NULL) exit(1);
 printf("UTILITY PROGRAM\n"); fflush(stdout);}

static boolean do_input()
{char A[4096]; fgets(A,4000,stdin); A[strlen(A)-1]=0;
 if (!strcmp(A,"quit")) return FALSE;
 if (!memcmp(A,"getinfo ",8)) do_getinfo(ROOT_TREE,RPOS,A+8);
 if (!memcmp(A,"directline ",11)) do_directline(ROOT_TREE,RPOS,A+11,TRUE);
 if (!strcmp(A,"prune")) do_prune(ROOT_TREE,RPOS);
 if (!strcmp(A,"recompute"))
   do_subtree(ROOT_TREE,RPOS,"","BOOKS/RECOMPUTED.book",TRUE,TRUE);
 if (!memcmp(A,"extend ",7)) do_extend(ROOT_TREE,RPOS,A+7);
 if (!memcmp(A,"search ",7)) do_search(RPOS,A+7);
 if (!memcmp(A,"goodbad ",7)) do_goodbad(ROOT_TREE,RPOS,A+8); 
 if (!memcmp(A,"subtree ",8))
   do_subtree(ROOT_TREE,RPOS,A+8,"BOOKS/SUBTREE.book",TRUE,TRUE);
 if (!memcmp(A,"loadbook ",9)) do_loadbook(RPOS,A+9);
 if (!strcmp(A,"savebook")) do_savebook(ROOT_TREE,"BOOKS/MODIFIED.book");
 if (!memcmp(A,"excludesetup ",13)) do_excludesetup(ROOT_TREE,RPOS,A+13);
 if (!memcmp(A,"subexclude ",11)) do_exclude(ROOT_TREE,RPOS,A+11,TRUE);
 if (!memcmp(A,"excludemoves ",13)) do_exclude(ROOT_TREE,RPOS,A+13,FALSE);
 if (!memcmp(A,"wintree ",8)) doit_wintree(ROOT_TREE,RPOS,A+8);
 if (!memcmp(A,"merge ",6)) doit_merge(ROOT_TREE,A+6);
 printf("Done with input\n"); fflush(stdout); return TRUE;}

int main(int c,char **I) {init_stuff(c,I); while (do_input()); return 0;}
